@@ -27,13 +27,11 @@ except:
     basestring = unicode = str
 
 import warnings
-try:
-    from StringIO import StringIO
-except ImportError:
-    from io import BytesIO as StringIO
+from io import BytesIO
 import xml.etree.ElementTree as et
 
 from .specs import SPECS, REVERSE_SPECS
+from .wcs_utils import get_cdelt_crota, get_cd
 
 
 def register_namespace(tag, uri):
@@ -364,18 +362,29 @@ class AVM(AVMContainer):
             return object.__getattr__(self, attribute)
 
     @classmethod
-    def from_image(cls, filename):
+    def from_image(cls, filename, xmp_packet_index=None):
         """
         Instantiate an AVM object from an existing image.
+
+        Parameters
+        ----------
+        filename : str
+            The filename of the image to read the AVM meta-data from
+        xmp_packet_index : int, optional
+            In cases where multiple XMP packets are present in the file, this
+            can be used to indicate which one to use. If not specified, this
+            defaults to the first XMP packet found.
         """
 
         # Get XMP data from file
-        xmp = extract_xmp(filename)
+        xmp = extract_xmp(filename, xmp_packet_index=xmp_packet_index)
+
+        print(xmp)
 
         # Extract XML
-        start = xmp.index("<?xpacket begin=")
-        start = xmp.index("?>", start) + 2
-        end = xmp.index("</x:xmpmeta>") + 12
+        start = xmp.index(b"<?xpacket begin=")
+        start = xmp.index(b"?>", start) + 2
+        end = xmp.index(b"</x:xmpmeta>") + 12
 
         # Extract XML
         xml = xmp[start:end]
@@ -397,7 +406,7 @@ class AVM(AVMContainer):
         self = cls()
 
         # Parse XML
-        tree = et.parse(StringIO(xml))
+        tree = et.parse(BytesIO(xml))
         root = tree.getroot()
         avm_content = parse_avm_content(root)
 
@@ -427,9 +436,26 @@ class AVM(AVMContainer):
 
         return self
 
-    def to_wcs(self, use_full_header=False):
+    def to_wcs(self, use_full_header=False, target_image=None, target_shape=None):
         """
-        Convert AVM projection information into a Astropy WCS object
+        Convert AVM projection information into a Astropy WCS object.
+
+        Parameters
+        ----------
+        use_full_header : bool, optional
+            Whether to use the full embedded Header if available. If set to
+            `False`, the WCS is determined from the regular AVM keywords.
+        target_image : str, optional
+            In some cases, the dimensions of the image containing the AVM/WCS
+            information is different from the dimensions of the image for which
+            the AVM was defined. The `target_image` option can be used to pass
+            the path of an image from which the size will be used to re-scale
+            the WCS.
+        target_shape : tuple of ``(nx: int, ny: int)``, optional
+            Serves the same function as `target_image`, but for cases where the
+            target dimensions are known without having to open and read an image
+            file. If both this and `target_image` are specified, `target_image`
+            takes precedence.
         """
 
         if not astropy_installed:
@@ -440,7 +466,7 @@ class AVM(AVMContainer):
 
         if use_full_header and self.Spatial.FITSheader is not None:
             print("Using full FITS header from Spatial.FITSheader")
-            header = fits.Header(txtfile=StringIO(self.Spatial.FITSheader))
+            header = fits.Header(txtfile=BytesIO(self.Spatial.FITSheader))
             return WCS(header)
 
         # Initializing WCS object
@@ -493,7 +519,13 @@ class AVM(AVMContainer):
             wcs.wcs.equinox = float(self.Spatial.Equinox)
 
         # Set standard WCS parameters
-        wcs.naxis1, wcs.naxis2 = self.Spatial.ReferenceDimension
+        if self.Spatial.ReferenceDimension is not None:
+            wcs_naxis1, wcs_naxis2 = self.Spatial.ReferenceDimension
+            if hasattr(wcs, 'naxis1'):  # PyWCS and Astropy < 0.4
+                wcs.naxis1, wcs.naxis2 = wcs_naxis1, wcs_naxis2
+        else:
+            wcs_naxis1, wcs_naxis2 = None, None
+
         wcs.wcs.crval = self.Spatial.ReferenceValue
         wcs.wcs.crpix = self.Spatial.ReferencePixel
 
@@ -501,10 +533,50 @@ class AVM(AVMContainer):
             wcs.wcs.cd = [self.Spatial.CDMatrix[0:2],
                           self.Spatial.CDMatrix[2:4]]
         elif self.Spatial.Scale is not None:
-            wcs.wcs.cdelt = self.Spatial.Scale
+            # AVM Standard 1.2:
+            #
+            # "The scale should follow the standard FITS convention for sky
+            # projections in which the first element is negative (indicating
+            # increasing RA/longitude to the left) and the second is positive.
+            # In practice, only the absolute value of the first term should be
+            # necessary to identify the pixel scale since images should always
+            # be presented in an undistorted 1:1 aspect ratio as they appear in
+            # the sky when viewed from Earth.This field can be populated from
+            # the FITS keywords: CDELT1, CDELT2 (or derived from CD matrix)."
+            #
+            # Therefore, we have to enforce the sign of CDELT:
+            wcs.wcs.cdelt[0] = - abs(self.Spatial.Scale[0])
+            wcs.wcs.cdelt[1] = + abs(self.Spatial.Scale[1])
             if self.Spatial.Rotation is not None:
                 wcs.wcs.crota = self.Spatial.Rotation, self.Spatial.Rotation
 
+        # If `target_image` is set, we have to rescale the reference pixel and
+        # the scale
+        if target_image is not None:
+            from PIL import Image
+            target_shape = Image.open(target_image).size
+
+        if target_shape is not None:
+            if self.Spatial.ReferenceDimension is None:
+                raise ValueError("Spatial.ReferenceDimension should be set in order to determine scale in target image")
+
+            # Find scale in x and y
+            scale_x = target_shape[0] / float(wcs_naxis1)
+            scale_y = target_shape[1] / float(wcs_naxis2)
+
+            # Check that scales are consistent
+            if abs(scale_x - scale_y) / (scale_x + scale_y) * 2. < 0.01:
+                scale = scale_x
+            else:
+                raise ValueError("Cannot scale WCS to target image consistently in x and y direction")
+
+            wcs.wcs.cdelt /= scale
+            wcs.wcs.crpix *= scale
+
+            if hasattr(wcs, 'naxis1'):  # PyWCS and Astropy < 0.4
+                wcs.naxis1 = target_shape[0]
+                wcs.naxis2 = target_shape[1]
+
         return wcs
 
     @classmethod
@@ -517,7 +589,8 @@ class AVM(AVMContainer):
             raise Exception("Astropy is required to use from_wcs()")
 
         wcs = WCS(header)
-        self = cls.from_wcs(wcs)
+        shape = (header['NAXIS2'], header['NAXIS1'])
+        self = cls.from_wcs(wcs, shape=shape)
 
         if include_full_header:
             self.Spatial.FITSheader = str(header)
@@ -525,9 +598,16 @@ class AVM(AVMContainer):
         return self
 
     @classmethod
-    def from_wcs(cls, wcs):
+    def from_wcs(cls, wcs, shape=None):
         """
         Instantiate an AVM object from a WCS transformation
+
+        Parameters
+        ----------
+        wcs : `~astropy.wcs.WCS` instance
+            The WCS to convert to AVM
+        shape : tuple, optional
+            The shape of the image (using Numpy y, x order)
         """
 
         if not astropy_installed:
@@ -548,10 +628,25 @@ class AVM(AVMContainer):
         else:
             raise Exception("Projections do not agree: %s / %s" % (proj1, proj2))
 
-        self.Spatial.ReferenceDimension = [wcs.naxis1, wcs.naxis2]
+        try:
+            self.Spatial.ReferenceDimension = [wcs.naxis1, wcs.naxis2]
+        except:
+            if shape is None:
+                warnings.warn("no shape specified, so Spatial.ReferenceDimension will not be set")
+            else:
+                self.Spatial.ReferenceDimension = [shape[1], shape[0]]
+
         self.Spatial.ReferenceValue = wcs.wcs.crval.tolist()
         self.Spatial.ReferencePixel = wcs.wcs.crpix.tolist()
-        self.Spatial.Scale = wcs.wcs.cdelt.tolist()
+
+        try:
+            cdelt1, cdelt2, crota2 = get_cdelt_crota(wcs)
+            self.Spatial.Scale = cdelt1, cdelt2
+            self.Spatial.Rotation = crota2
+        except ValueError:
+            warnings.warn("cannot represent WCS as a Scale and Rotation, using deprecated CDMatrix instead")
+            cd = get_cd(wcs)
+            self.Spatial.CDMatrix = cd
 
         xcoord = decode_ascii(wcs.wcs.ctype[0][:4])
         ycoord = decode_ascii(wcs.wcs.ctype[1][:4])
@@ -571,11 +666,6 @@ class AVM(AVMContainer):
         else:
             raise Exception("Unknown coordinate system: {0}/{1}".format(xcoord, ycoord))
 
-        try:
-            self.Spatial.Rotation = wcs.wcs.crota[1]
-        except:
-            pass
-
         self.Spatial.Quality = "Full"
 
         return self
@@ -617,8 +707,8 @@ class AVM(AVMContainer):
         # Create XML Tree
         tree = et.ElementTree(root)
 
-        # Need to create a StringIO object to write to
-        s = StringIO()
+        # Need to create a BytesIO object to write to
+        s = BytesIO()
         tree.write(s, encoding='utf-8')
 
         # Rewind and read the contents
